import * as packageJson from '#package.json' with { type: 'json' };
import { kimiEvent } from '@nldoc/logger';
import { connect } from 'amqplib';
import { Client as MinioClient } from 'minio';
import { createHash, randomUUID } from 'node:crypto';
import { ConnectionError, ExternalFileStorageNotReadyError, FailedQueueAssertion, InvalidJob, InvalidWorkerScout, WorkerError, } from './errors.js';
import { MinioRemoteFileStorage } from './MinioRemoteFileStorage.js';
import { WorkerAvailabilityStatus } from './types/WorkerAvailability.js';
import { WorkerJob } from './types/WorkerJob.js';
import { WorkerScout } from './types/WorkerScout.js';
import { sleep } from './utils/index.js';
export class Worker {
    constructor({ logger, messageHandler, config, traceId, }) {
        this.jobAckRoutingKeyPrefix = `ack.`;
        this.jobPattern = 'job.*';
        this.jobResultRoutingKeyPrefix = 'result';
        this.stationWorkerScoutRoutingKey = 'station.workers';
        this.minioUseSSL = false;
        this.inFlight = new Set();
        this.shuttingDown = false;
        this.heartbeatIntervalSeconds = 60;
        this.isRestarting = false;
        this.onTerminationSignal = async (signal) => {
            this.logger.info(`${signal} signal received.`);
            return this.shutdown();
        };
        this.shutdown = async () => {
            this.logger.info(`Gracefully shutting down ${this.workerName}`);
            this.logger.info(`Shutting down connection...`);
            this.shuttingDown = true;
            clearInterval(this.availabilityInterval);
            this.logger.info(`Setting channel prefetch to 0`);
            await this.amqpChannel?.prefetch(0, false);
            await this.inFlightDrained();
            this.logger.info('Closing AMQP client connection');
            await this.amqpClient?.close();
            this.logger.info(`Shutting down!`);
            this.logger.flush();
            await sleep(1000);
            process.exit();
        };
        this.handleConnectionError = (err) => {
            if (this.shuttingDown) {
                return;
            }
            this.logger.warn({ err }, 'A connection error occurred. Restarting connection...');
            void this.restart();
        };
        this.handleConnectionClose = () => {
            if (this.shuttingDown) {
                return;
            }
            this.logger.warn('A connection was unexpectedly closed. Restarting connection...');
            void this.restart();
        };
        this.workerInstanceName = config.workerInstanceName;
        this.workerName = config.name;
        this.workerType = config.exchangePrefix;
        this.amqpHost = config.amqpHost;
        this.amqpPass = config.amqpPass;
        this.amqpPort = config.amqpPort;
        this.amqpProtocol = config.amqpProtocol;
        this.amqpUser = config.amqpUser;
        this.prefetchLimit = config.prefetchLimit ?? 10;
        this.minioHost = config.minioHost;
        this.minioPort = config.minioPort;
        this.minioAccessKey = config.minioAccessKey;
        this.minioSecretKey = config.minioSecretKey;
        this.minioOutputBucketName = config.minioOutputBucketName;
        this.minioDebugBucketName = config.minioDebugBucketName;
        this.minioUseSSL = config.minioUseSSL ?? false;
        this.exchangePrefix = config.exchangePrefix;
        this.exchangeJobs = `${this.exchangePrefix}-jobs`;
        this.exchangeHealth = `${this.exchangePrefix}-workers`;
        this.exchangeResults = `${this.exchangePrefix}-job-results`;
        this.queueName = this.workerName;
        this.instanceTraceId = traceId ?? randomUUID();
        this.logger = logger.child({
            trace: { id: this.instanceTraceId },
            worker: {
                name: this.workerName,
                type: this.workerType,
                instanceName: this.workerInstanceName,
                baseWorkerVersion: packageJson.default.version,
            },
        });
        if (!this.workerName.startsWith('worker-')) {
            this.logger.warn(`Worker name "${this.workerName}" does not implement naming convention. Please prefix the workerName with "worker-"`);
        }
        this.messageHandler = messageHandler;
    }
    getInfo() {
        return {
            amqpHost: this.amqpHost,
            amqpPort: this.amqpPort,
            amqpProtocol: this.amqpProtocol,
            prefetchLimit: this.prefetchLimit,
            exchangeJobs: this.exchangeJobs,
            exchangeHealth: this.exchangeHealth,
            exchangeResults: this.exchangeResults,
            queue: this.queueName,
            workerInstance: this.workerInstanceName,
            workerName: this.workerName,
            workerType: this.workerType,
        };
    }
    async start() {
        this.attachExternalEventListeners();
        await this.connectMinio();
        await this.connect();
        await this.setupAmqpChannel();
        this.startHeartbeat();
        await this.consumeChannels();
    }
    async restart() {
        if (this.isRestarting) {
            this.logger.info('Restart already in progress, ignoring duplicate request');
            return;
        }
        this.isRestarting = true;
        this.logger.info('Restarting worker connection');
        clearInterval(this.availabilityInterval);
        try {
            if (this.amqpChannel) {
                try {
                    this.logger.info('Closing channel before restart');
                    await this.amqpChannel.close();
                }
                catch (err) {
                    this.logger.warn({ err }, `Error closing channel`);
                }
            }
            if (this.amqpClient) {
                try {
                    this.logger.info('Closing connection before restart');
                    await this.amqpClient.close();
                }
                catch (err) {
                    this.logger.warn({ err }, `Error closing connection`);
                }
            }
        }
        catch (err) {
            this.logger.warn({ err }, `Error during connection cleanup`);
        }
        this.amqpChannel = undefined;
        this.amqpClient = undefined;
        await sleep(1000);
        this.logger.info('Starting new connection');
        try {
            await this.start();
        }
        finally {
            this.isRestarting = false;
        }
    }
    attachExternalEventListeners() {
        this.logger.info('Attach SIGTERM and SIGINT listeners');
        process.once('SIGTERM', this.onTerminationSignal);
        process.once('SIGINT', this.onTerminationSignal);
    }
    async inFlightDrained() {
        while (this.inFlight.size > 0) {
            this.logger.info(`There are ${this.inFlight.size} messages in flight...`);
            await sleep(200);
        }
    }
    async connectMinio() {
        if (!this.minioHost || !this.minioAccessKey || !this.minioOutputBucketName) {
            const debugObject = {
                minioHost: this.minioHost,
                minioAccessKey: this.minioAccessKey,
                minioOutputBucketName: this.minioOutputBucketName,
            };
            throw new ConnectionError(`MinIO failed to connect, as one of the connection parameters is empty: ${JSON.stringify(debugObject)}`);
        }
        if (!this.minioSecretKey) {
            throw new ConnectionError('MinIO failed to connect, as the minioSecretKey is not defined');
        }
        this.minioClient = new MinioClient({
            endPoint: this.minioHost,
            port: this.minioPort,
            useSSL: this.minioUseSSL,
            accessKey: this.minioAccessKey,
            secretKey: this.minioSecretKey,
        });
        const bucketExists = await this.minioClient.bucketExists(this.minioOutputBucketName);
        if (!bucketExists) {
            throw new Error(`MinIO output bucket ${this.minioOutputBucketName} does not exist.`);
        }
    }
    async connect() {
        this.logger.info(`Connecting to host: ${this.amqpHost} (${this.amqpProtocol}://${this.amqpUser}@${this.amqpHost}:${this.amqpPort})`);
        let connectionRetries = 0;
        while (!this.shuttingDown && !this.amqpClient) {
            try {
                this.amqpClient = await connect(`${this.amqpProtocol}://${this.amqpUser && this.amqpPass ? `${this.amqpUser}:${this.amqpPass}@` : ''}${this.amqpHost}:${this.amqpPort}`);
            }
            catch (err) {
                connectionRetries++;
                this.logger[connectionRetries > 30 ? 'warn' : 'debug'](`Failed to connect... Retrying (${connectionRetries})`);
                await sleep(2000);
            }
        }
        this.amqpClient?.on('error', this.handleConnectionError);
        this.amqpClient?.on('close', this.handleConnectionClose);
        this.logger.info(`Connected to host: ${this.amqpHost} (${this.amqpProtocol}://${this.amqpHost}:${this.amqpPort})`);
    }
    async setupChannel() {
        if (this.amqpChannel) {
            try {
                this.logger.info('Cleaning up existing channel');
                await this.amqpChannel.close();
            }
            catch (err) {
                this.logger.warn({ err }, `Error closing channel`);
            }
            this.amqpChannel = undefined;
        }
        await this.assertChannel();
    }
    async setupAmqpChannel() {
        await this.setupChannel();
        await this.assertExchanges();
        await this.assertQueues();
        await this.bindQueues();
    }
    async assertChannel() {
        this.logger.info(`Assert channel`);
        this.amqpChannel = await this.amqpClient?.createChannel();
        await this.amqpChannel?.prefetch(this.prefetchLimit, false);
        this.amqpChannel?.on('error', this.handleConnectionError);
        this.amqpChannel?.on('close', this.handleConnectionClose);
        this.logger.info(`Asserted channel`);
    }
    async assertExchanges() {
        return Promise.all([
            this.assertExchange(this.exchangeHealth),
            this.assertExchange(this.exchangeJobs),
            this.assertExchange(this.exchangeResults),
        ]);
    }
    async assertExchange(exchangeName) {
        this.logger.info(`Assert exchange: ${exchangeName}`);
        await this.amqpChannel?.assertExchange(exchangeName, 'fanout', { durable: true });
        this.logger.info(`Asserted exchange: ${exchangeName}`);
    }
    async assertQueues() {
        await this.assertQueue({ name: this.queueName, type: 'quorum' });
        this.healthQueueName = await this.assertExclusiveQueue();
    }
    async assertExclusiveQueue() {
        const result = await this.amqpChannel?.assertQueue('', { exclusive: true });
        const asserted_name = result?.queue;
        if (asserted_name == null) {
            this.logger
                .child({ data: { assertExclusiveQueueResult: result } })
                .error(`Failed to create exclusive queue`);
            throw new FailedQueueAssertion(`Failed to create exclusive queue`);
        }
        this.logger.info(`Asserted exclusive queue: ${asserted_name}`);
        return asserted_name;
    }
    async assertQueue({ name, type = 'quorum' }) {
        this.logger.info(`Assert ${type} queue with name: '${name}'`);
        const result = await this.amqpChannel?.assertQueue(name ? name : '', {
            exclusive: false,
            arguments: { 'x-queue-type': type },
        });
        const asserted_name = result?.queue;
        if (asserted_name == null) {
            this.logger.error('Failed to create queue');
        }
        else {
            this.logger.info(`Asserted queue: ${asserted_name}`);
        }
        return asserted_name;
    }
    async bindQueues() {
        var operations = [
            this.bindQueue({
                queueName: this.queueName,
                exchangeName: this.exchangeJobs,
                routingKey: this.jobPattern,
            }),
        ];
        if (this.healthQueueName) {
            operations.push(this.bindQueue({
                queueName: this.healthQueueName,
                exchangeName: this.exchangeHealth,
                routingKey: this.stationWorkerScoutRoutingKey,
            }));
        }
        else {
            this.logger.warn('No health queue name found, did not bind health queue to exchange');
        }
        return Promise.all(operations);
    }
    async bindQueue({ queueName, exchangeName, routingKey, }) {
        this.logger.info(`Binding queue ${queueName} to exchange ${exchangeName} with pattern ${routingKey}`);
        await this.amqpChannel?.bindQueue(queueName, exchangeName, routingKey);
        this.logger.info(`Bound queue ${queueName} to exchange ${exchangeName}`);
    }
    publish({ exchange, routingKey, body, logger, traceId, logLevel = 'info', }) {
        if (!this.amqpChannel) {
            logger.warn('Could not publish message, channel not available');
            return false;
        }
        try {
            logger[logLevel](`Publishing message to ${exchange}/${routingKey}`);
            return this.amqpChannel.publish(exchange, routingKey, Buffer.from(JSON.stringify(body), 'utf8'), {
                headers: {
                    'x-kimi-worker-type': this.workerType,
                    'x-kimi-worker-name': this.workerName,
                    'x-kimi-worker-instance-name': this.workerInstanceName,
                    'x-trace-id': traceId,
                },
                contentType: 'application/json',
                contentEncoding: 'utf8',
            });
        }
        catch (err) {
            const errorMessage = err instanceof Error ? err.message : String(err);
            logger.warn(`Error publishing message: ${errorMessage}`);
            if (err instanceof Error &&
                (errorMessage.includes('Channel closed') ||
                    errorMessage.includes('connection') ||
                    errorMessage.includes('socket'))) {
                void this.handleConnectionError(err);
                return false;
            }
            else {
                logger.error({ err }, `Unexpected error publishing message: ${errorMessage}`);
                return false;
            }
        }
    }
    sendAvailability({ state, reason, logger, traceId, }) {
        const routingKey = `worker.${this.workerName}.${this.workerInstanceName}`;
        logger.trace(`Sending state: ${state} (${reason}) -> ${this.exchangeResults}/${routingKey}`);
        const body = {
            workerType: this.workerType,
            workerName: this.workerName,
            workerInstance: this.workerInstanceName,
            state,
            reason,
            timestamp: Date.now().toString(),
        };
        const success = this.publish({
            exchange: this.exchangeResults,
            routingKey,
            body,
            traceId,
            logger,
            logLevel: 'trace',
        });
        if (!success) {
            logger.warn('Failed to send worker availability status');
        }
    }
    startHeartbeat() {
        if (this.availabilityInterval) {
            clearInterval(this.availabilityInterval);
        }
        const availabilityTraceArgs = {
            traceId: this.instanceTraceId,
            logger: this.logger.child({ trace: { id: this.instanceTraceId } }),
        };
        try {
            this.sendAvailability({
                state: WorkerAvailabilityStatus.enum.Active,
                reason: 'Worker started',
                ...availabilityTraceArgs,
            });
        }
        catch (err) {
            this.logger.warn({ err }, `Failed to send initial availability status`);
        }
        this.logger.info(`Starting heartbeat interval, every ${this.heartbeatIntervalSeconds} seconds.`);
        this.availabilityInterval = setInterval(() => {
            try {
                this.heartbeat(availabilityTraceArgs);
            }
            catch (err) {
                this.logger.warn({ err }, `Error sending heartbeat`);
            }
        }, this.heartbeatIntervalSeconds * 1000);
    }
    heartbeat({ traceId, logger }) {
        this.sendAvailability({
            state: WorkerAvailabilityStatus.enum.Active,
            reason: 'heartbeat',
            traceId,
            logger,
        });
    }
    sendJobAck({ traceId, recordId, jobId, logger, }) {
        logger.info(`Sending job ack`);
        const body = { traceId, recordId, jobId };
        const routingKey = `${this.jobAckRoutingKeyPrefix}${jobId}`;
        const success = this.publish({
            exchange: this.exchangeResults,
            routingKey,
            body,
            logger,
            traceId,
        });
        if (!success) {
            logger.warn('Failed to send worker job ack');
        }
    }
    checkJobMessageRoutingKey(msg) {
        if (msg.fields.routingKey.split('.')[0] !== this.jobPattern.split('.')[0]) {
            throw new InvalidJob(`Received message that doesn't follow pattern "${this.jobPattern}": ${msg.fields.routingKey}`);
        }
    }
    parseJobMessage(msg) {
        const job = WorkerJob.safeParse(JSON.parse(msg.content.toString()));
        if (!job.success) {
            throw new InvalidJob('Job content could not be parsed', job.error);
        }
        const jobId = createHash('md5')
            .update(`${msg.properties.headers?.['x-trace-id'] ?? randomUUID()}${job.data.recordId}${this.workerName}`)
            .digest('hex');
        return { job: job.data, jobId };
    }
    async publishResult({ job, jobId, result, traceId, logger, }) {
        const routingKey = `${this.jobResultRoutingKeyPrefix}.${jobId}`;
        const body = {
            traceId,
            recordId: job.recordId,
            jobId,
            timestamp: new Date().toISOString(),
            ...result,
        };
        return this.publish({ exchange: this.exchangeResults, routingKey, body, traceId, logger });
    }
    async consumeChannels() {
        this.logger.info(`Start consuming (${this.queueName})`);
        const timestamp = Date.now();
        const consumerTag = `${process.env.HOSTNAME}-${timestamp}`;
        const healthConsumerTag = `${process.env.HOSTNAME}-health-${timestamp}`;
        this.logger.info(`Using consumer tag: ${consumerTag}`);
        await this.amqpChannel?.consume(this.queueName, (msg) => this.handleMessage(msg), { consumerTag });
        if (this.healthQueueName) {
            this.logger.info(`Using health consumer tag: ${healthConsumerTag}`);
            await this.amqpChannel?.consume(this.healthQueueName, (msg) => this.handleMessage(msg), { consumerTag: healthConsumerTag });
        }
        else {
            this.logger.warn('No health queue name found, did not start consuming health queue');
        }
    }
    handleMessage(msg) {
        if (!msg) {
            this.logger.trace(`Received empty message...`);
            return;
        }
        const traceId = msg.properties.headers?.['x-trace-id'] ?? randomUUID();
        const logger = this.logger.child({ trace: { id: traceId } });
        if (this.shuttingDown) {
            logger.info(`Message incoming after shutdown signal, reschedule message on queue (nack: ${msg.fields.deliveryTag})`);
            this.amqpChannel?.nack(msg, false, true);
            return;
        }
        void this.handleMessageAsync(msg, logger, traceId);
    }
    async handleMessageAsync(msg, logger, traceId) {
        let job = null;
        let jobId = null;
        try {
            this.inFlight.add(msg.fields.deliveryTag);
            if (msg.properties.headers?.['x-kimi-worker-name'] === this.workerName) {
                logger.trace(`Received message from self... Ignore that...`);
                return;
            }
            if (msg.fields.routingKey === this.stationWorkerScoutRoutingKey) {
                await this.handleWorkerScoutMessage(msg, logger, traceId);
                return;
            }
            if (!this.minioClient) {
                throw new ExternalFileStorageNotReadyError('MinIO client not initialized when handling the job');
            }
            this.checkJobMessageRoutingKey(msg);
            ({ job, jobId } = this.parseJobMessage(msg));
            this.logger.child({ data: { fields: msg.fields, jobId } }).info('Incoming job');
            const jobLogger = logger.child({ jobId, recordId: job.recordId });
            this.sendJobAck({ traceId, recordId: job.recordId, jobId, logger: jobLogger });
            await this.storeJobInMinio(job, jobId, traceId, jobLogger);
            const startTime = new Date();
            const result = await this.processJob(job, jobId, jobLogger);
            kimiEvent(jobLogger, {
                name: 'workerJobCompleted',
                eventStartTime: startTime,
                eventTime: new Date(),
            });
            if (!(await this.publishResult({ job, jobId, result, logger: jobLogger, traceId }))) {
                throw new WorkerError('Handler result could not be published');
            }
        }
        catch (error) {
            await this.handleProcessingError(error, job, jobId, logger, traceId);
        }
        finally {
            this.amqpChannel?.ack(msg);
            this.inFlight.delete(msg.fields.deliveryTag);
        }
    }
    async handleWorkerScoutMessage(msg, logger, traceId) {
        const jsonContent = this.tryParseJson(msg.content.toString());
        const workerScoutParseResult = WorkerScout.safeParse(jsonContent);
        if (!workerScoutParseResult.success) {
            throw new InvalidWorkerScout('WorkerScout content could not be parsed', workerScoutParseResult.error);
        }
        this.sendAvailability({
            state: WorkerAvailabilityStatus.enum.Active,
            reason: `Answering Worker Scout: ${workerScoutParseResult.data.reason}`,
            traceId,
            logger,
        });
    }
    async processJob(job, jobId, logger) {
        return this.messageHandler({
            logger: logger.child({ data: { jobId } }),
            job,
            jobId,
            storage: new MinioRemoteFileStorage({ minio: this.minioClient }), // We check for minioClient before
        });
    }
    async handleProcessingError(error, job, jobId, logger, traceId) {
        const errorMessage = error instanceof Error ? error.message : 'An unknown error occurred';
        logger.error(error instanceof Error ? error : errorMessage);
        if (job && jobId) {
            const errorResult = {
                success: false,
                confidence: 0,
                stringResult: errorMessage,
                error: {
                    code: 0, // TODO: refactor to match new event structure
                    message: errorMessage,
                },
            };
            try {
                const published = await this.publishResult({
                    job,
                    jobId,
                    result: errorResult,
                    logger,
                    traceId,
                });
                if (published) {
                    kimiEvent(logger, { name: 'workerJobFailed', eventTime: new Date() });
                }
                else {
                    logger.error('Failed to publish error result - publish returned false');
                }
            }
            catch (publishError) {
                logger.error(`Failed to publish error result: ${publishError instanceof Error ? publishError.message : String(publishError)}`);
            }
        }
    }
    tryParseJson(content) {
        if (!content) {
            return undefined;
        }
        try {
            return JSON.parse(content);
        }
        catch {
            this.logger.warn('Found non-json content in message');
            return undefined;
        }
    }
    async storeJobInMinio(job, jobId, traceId, logger) {
        const bucketName = this.minioDebugBucketName;
        if (!bucketName)
            return;
        const folderName = job.recordId;
        const fileName = `${jobId}.json`;
        const objectName = `${folderName}/${fileName}`;
        logger
            .child({ bucket_name: bucketName, folder_name: folderName, file_name: fileName })
            .debug('Storing job debug file in MinIO...');
        try {
            await this.minioClient?.putObject(bucketName, objectName, Buffer.from(JSON.stringify(job)), undefined, {
                jobId: jobId,
                recordId: job.recordId,
                worker: this.workerName,
                instance: this.instanceTraceId,
                traceId: traceId,
            });
        }
        catch (error) {
            if (error instanceof Error) {
                throw new ExternalFileStorageNotReadyError(`Could not store debug file ${fileName} to bucket ${bucketName}/${folderName}`, error);
            }
            else {
                throw error;
            }
        }
    }
}
//# sourceMappingURL=Worker.js.map